;*****************************************************************
;
;YING-ZONG HUANG
;December, 2002
;Wireless Freeform Input Device
;Receiver Code
;
;*****************************************************************

.nolist
.include "m323def.inc"
.list

; define constants
.equ	BEL				=0x07
.equ	BS				=0x08
.equ    CR      		=0x0D
.equ    LF      		=0x0A
.equ	SPACE			=0x20
.equ	ASC_M			=0x4D
.equ	ASC_2			=0x32

.equ	BAUD_RATE		=1200
.equ	CPU_CLOCK		=8000000
.equ	UART_BAUD_DIV	=(CPU_CLOCK/16/BAUD_RATE)-1

.equ    TMR0_PRESCL	=0x04	; this is the prescaler setting for CK/256	= ~56 OVFL interrupts per second
.equ	TMR0_INT_EN =1<<TOIE0		; set the timer 0 ovfl intr enable bit
.equ	TMR0_INT_DS	=~TMR0_INT_EN
.equ	TMR1_PRESCL =0x01	; no scale
.equ	TMR1_PWM_EN =0x83	; 10 bits, non-inverted up down PWM EN
.equ	TMR1_PWM_DS	=0x80	; PWM disable...
.equ	TMR1_HALFH =0x01	; 0x01FF
.equ	TMR1_HALFL =0xFF

.equ	BUFSIZE		=100
.equ	QSIZE		=BUFSIZE+1

; define register aliases
.def	pr		=r0
.def	zero	=r1
.def	all		=r2
.def	a1		=r3		; function argument 1
.def	v0		=r4		; function return 0
.def	v1		=r5		; function return 1
.def	t0		=r16	; temporary register 0h
.def	t1		=r17	; temporary register 1
.def	t2		=r18
.def	t3		=r19
.def	s0		=r20
.def	s1		=r21
.def	s2		=r22
.def	s3		=r23
.def	a0		=r26		; function argument 0
.def    qHead		=r24	; index of beginning of queue buffer
.def    qTail		=r25	; index of end of queue buffer
.def	z		=r30

; data segment
        .dseg
qbuf:	.byte	QSIZE	; define an array of bytes qbuf[QSIZE]


; code segment	 interrupt table must use the 2-byte 'jmp'
        .cseg
        .org    $000
	jmp		reset		; $000 HW reset or watchdog
	jmp		reset		; External Interrupt0 Vector Address
	jmp		reset		; External Interrupt1 Vector Address
	jmp		reset		; External Interrupt2 Vector Address
	jmp		reset		; Output Compare2 Interrupt Vector Address
	jmp		reset		; Overflow2 Interrupt Vector Address
	jmp		reset		; Input Capture1 Interrupt Vector Address
	jmp		reset		; Output Compare1A Interrupt Vector Address
	jmp		reset		; Output Compare1B Interrupt Vector Address
	jmp		reset		; Overflow1 Interrupt Vector Address
	jmp		reset		; Output Compare0 Interrupt Vector Address
	jmp		reset		; Overflow0 Interrupt Vector Address
	jmp		reset		; SPI Interrupt Vector Address
	jmp		RxC_Isr		; USART Receive Complete Interrupt Vector Address
	jmp		reset		; USART Data Register Empty Interrupt Vector Address
	jmp		reset		; USART Transmit Complete Interrupt Vector Address
	jmp		reset		; ADC Interrupt Vector Address
	jmp		reset		; EEPROM Interrupt Vector Address
	jmp		reset		; Analog Comparator Interrupt Vector Address
	jmp		reset		; Irq. vector address for Two-Wire Interface
	jmp		reset		; Store Program Memory Ready Interrupt Vector Address

; begin main code
reset:
	ldi 	t0, low(RAMEND)		; initialize stack pointer to highest (valid) mem location
	out		SPL, t0
	ldi 	t0, high(RAMEND)
	out		SPH, t0

	clr		zero		; set zero register to actually zero
	clr		all
	com		all
	rcall   clrqueue				; initialize queue

	ldi		t0, $7F				; port D all outputs, except D0 is input (RXD)
	out		DDRD, t0

	; Set baud rate
	ldi	t0, low(UART_BAUD_DIV)
	out UBRRL, t0
	ldi	t0, high(UART_BAUD_DIV) ;send at 1200 baud
	out UBRRH, t0
	ldi	t0, (0<<U2X)	; non double speed mode
	out	UCSRA, t0
	; Enable Receiver and Transmitter, receive interrupt
	ldi t0, (1<<RXEN)|(1<<TXEN)|(1<<RXCIE)
	out UCSRB,t0
	; Set frame format: 7data, 1stop bit
	ldi t0, (1<<URSEL)|(0<<USBS)|(2<<UCSZ0)
	out UCSRC,t0

	sei                             ; enable interrupts

loopchar:
	rcall   getchar				; get next character from queue
	mov		a0, v0
	rcall	putchar
	rjmp    loopchar


; begin subroutines and functions

; function getchar
; get character from receive serial queue
; if none available, wait until one arrives
getchar:
	rcall   dequeue
	brcc    getchar	; try again if queue empty
	ret

; function putchar
; output character in a0 to serial port, non-interrupt driven
putchar:
	sbis    UCSRA, UDRE			; loop until USR:UDRE is 1
	rjmp    putchar
	out     UDR, a0			; write a0 to transmitter buffer
	ret

; Interrupt Handler RxC_Isr
; UART receive character interrupt service routine
RxC_Isr:
	push    t0				; save registers
	in      t0, SREG
	push	t0
	push	a0
        
	in      a0, UDR			; read UART receive data
	rcall   enqueue				; place data in the queue

	pop		a0				; restore registers
	pop		t0
	out     SREG, t0                
	pop     t0
	reti						; return from interrupt


; function clrqueue
; initialize the queue buffer to empty (beginning=end=0)
clrqueue:
	clr     qHead				; make pointers equal
	clr     qTail
	ret

; function enqueue
; add byte in a0 to end of queue buffer
enqueue:
	push    yh					; save registers
	push    yl
	push	t0					; save this if it is possible to get interrupt in an ISR???
        
	ldi     yl, low(qbuf)		; y = base of queue
	ldi     yh, high(qbuf)
	add     yl, qHead			; add offset of head
	adc     yh, zero			; to get to end of queue
	st      y, a0				; store data in queue

	mov		t0, qHead
	inc		t0
	cpi		t0, QSIZE
	brlo	enq1
	clr		t0
enq1:
	cp		t0, qTail			; compare head to tail
	breq	enq2				; if head equals tail, we ain't incrementing the head pointer
	mov		qHead, t0			; queue has room, back-copy incremented pointer
enq2:
	pop		t0
	pop     yl
	pop     yh
	ret				; return

; function enqueue
; remove byte from beginning of queue and return in v0
; if data was available, carry is set
; if no data was available, carry is clear
dequeue:
	cp      qHead, qTail            ; check for empty queue
	brne    deq1					; queue not empty, actually dequeue at deq1...
	clc                             ; clear carry flag when empty
	ret
deq1:
	push    yh                      ; save queue base pointers
	push    yl
	ldi     yl, low(qbuf)           ; y = base of queue
	ldi     yh, high(qbuf)
	add     yl, qTail               ; add offset to get to tail
	adc     yh, zero				; the tail of the queue
	ld      v0, y					; fetch current tail element in queue
	inc     qTail					; update tail pointer
	cpi     qTail, QSIZE
	brlo    deq2
	clr     qTail                   ; wraparound to 0 when needed
deq2:
	pop     yl                      ; restore registers
	pop     yh
	sec                             ; set carry to indicate data
	ret
